from typing import Any, Callable

def compose(*funcs: Callable[..., Any]) -> Callable[..., Any]: ...
def compound(mode: str) -> Callable[[int], int]: ...
def simple(mode: str) -> Callable[[int], int]: ...
